from os.path import join
from PIL import Image
from discord.enums import ChannelType
from discord import ButtonStyle, Embed, Intents, Member, Interaction, Message, TextChannel, VoiceChannel
from discord.abc import GuildChannel
from discord.ext.commands import Command, Bot, Context
from sys import argv, path, stdin
from itertools import product
from asyncio import get_running_loop, run, StreamReader, StreamReaderProtocol, create_task
from typing import Callable, Any
import threading

from .components import *
from .card import Card
from .object import Object
from .colors import *
from .font import Font as CFFont
from .vector2 import Vector2
from .user import User
from .data import Data
from .font import Font

import logging

class Cord(Bot):
    def __init__(_, entry_command: str = None, autosave: bool = True) -> None:
        _.logger = logging.getLogger("CordForge")
        _.logger.setLevel(logging.DEBUG)  # or INFO
        if not _.logger.hasHandlers():
            handler = logging.StreamHandler()  # default to stdout
            formatter = logging.Formatter("[%(asctime)s] %(levelname)s: %(message)s")
            handler.setFormatter(formatter)
            _.logger.addHandler(handler)
        _.entry_command = entry_command
        _.entry:Callable = None
        _.setup:Callable = None
        _.autosave = autosave
        _.source_directory = path[0]
        _.instance_user:str = argv[1]
        _.user_dashboards:dict[str:Panel] = {}
        _.user_traits:list[list[str, Any]] = []
        _.user_profiles = {}
        _.message:Message = None
        _.font:Font = None
        _.objects:dict[str:Object] = {}
        _.data = Data(_)

        _._handle_alias() # required before handing _.prefix to Bot.__init__()
        super().__init__(command_prefix=_.prefix, intents=Intents.all())
        threading.Thread(target=_.sync_stdin_listener, args=(), daemon=True).start()
        

    def __setattr__(_, name, value):
        return super().__setattr__(name, value)


    def _handle_alias(_) -> None:
        _.prefix = [_.entry_command[0]]
        for prefix in _.prefix.copy():
            _.prefix.extend([variant for variant in _._all_case_variants(prefix, _.prefix)\
                                        if variant not in _.prefix])
        _.entry_command = [_.entry_command[1:]] 
        for alias in _.entry_command.copy():
            _.entry_command.extend([variant for variant in _._all_case_variants(alias, _.entry_command)\
                                        if variant not in _.entry_command])


    def _all_case_variants(_, string: str, originals:list[str]):
        pools = [(character.lower(), character.upper()) for character in string]
        variants = []
        for variant in product(*pools):
            string = ''.join(variant)
            if string not in originals: variants.append(string)
        return variants


    def _get_token(_, key:str) -> str:
        with open(join(_.source_directory, "keys")) as key_file:
            for line in key_file:
                line_data = line.split("=")
                if key.lower() == line_data[0].lower():
                    return line_data[1].strip()
        return "Could Not Find Token"
    

    def _setup_user_traits(_) -> None:
        for [trait, value] in _.user_traits:
            User.add_trait(trait, value)

            
    def sync_stdin_listener(_):
        print("Listening for input from launcher...")
        for line in stdin:
            line = line.strip()
            _.logger.info(f"[Launcher input]: {line}")


    def determine(_, dictionary:dict, callable:Callable=None) -> Any:
        '''
        Determines with key is being used, and selects the dependant items,\
        can optionally provide callback to is on determine.
        '''
        selection = dictionary[_.instance_user]
        if callable:
            selection = callable(selection)
        return selection
        

    def run_task(_, Task, *Arguments) -> Any:
        try: get_running_loop()
        except RuntimeError: return run(Task(*Arguments))
        raise RuntimeError("There is an existing loop. Run() is only used for setup before the Bot runs it's loop.")


    def launch(_, entry:Callable, setup:Callable=None) -> None:
        _.logger.info("Launching...")
        'Start Discord Bot'
        _.entry = entry
        _.setup = setup
        _.run(_._get_token(_.instance_user))


    async def setup_hook(_):
        async def wrapper(initial_context): await _._send_initial_card(initial_context)
        _.add_command(Command(wrapper, aliases=_.entry_command))
        await super().setup_hook()


    async def on_ready(_):
        _.logger.info("Bot is alive, and ready")
        _.logger.info("Setting up...")
        _._setup_user_traits()
        if _.setup: await _.setup()
        _.logger.info("Finished setup")
        await _.data.load_data()
        if _.autosave: await _.data.autosave()


    async def _send_initial_card(_, initial_context:Context) -> None:
        user:User = User(initial_context.author)
        user_card:Card = await _.new_card(user)

        try:
            await _.entry(user_card)
        except Exception as e:
            _.logger.info(f"Exception: {e}")

        if user.id not in _.user_profiles.keys():
            _.user_profiles.update({user.id:user})

        # Delete initial user message
        await initial_context.message.delete()

        # Delete past message, to avoid duplications
        if _.message is not None: await _.message.delete()

        # Construct Card, View, and Embed, and then send
        await user_card._construct()
        if user_card.view_frame.total_children_count > 0 and user_card.image == None:
            user_card.message = await initial_context.send(embed=user_card.embed_frame,
                                                           view=user_card.view_frame)
        elif user_card.image != None:
            user_card.embed_frame = Embed(title="")
            user_card.embed_frame.set_image(url="attachment://GameImage.png")
            await user_card._buffer_image()
            user_card.message = await initial_context.send(embed=user_card.embed_frame,
                                                           view=user_card.view_frame,
                                                           file=user_card.image_file)
        else:
            _.logger.info("Dashboard has nothing on it.")


    async def announce(_, channel:GuildChannel, message:str=None, card:Card=None) -> Message:
        '''
        Send a message, or Card to a specific channel
        '''
        if channel.type != ChannelType.text:_.logger.info("announce() is only compatible with TextChannels")
        else:\
        channel:TextChannel = channel
        if message and not card:
            await channel.send(message)
        elif message and card:
            card.embed_frame = Embed(title="")
            card.embed_frame.set_image(url="attachment://GameImage.png")
            await card._construct()
            await card._buffer_image()
            await channel.send(message, embed=card.embed_frame,
                               view=card.view_frame,
                               file=card.image_file)
            

    async def new_card(_, user:Member=None) -> Card:
        '''
        Create new card to draw on.

        Returns instantiated Card
        '''
        user_card:Card = Card(user)
        return user_card


    def load_image(_, image_path:str) -> Image:
        'Load image from file path into memory'
        return Image.open(image_path)


    async def reply(_, user_card:Card, interaction:Interaction) -> None:
        await user_card._construct()
        if user_card.view_frame.total_children_count > 0 and user_card.image == None:
            user_card.message = await interaction.response.edit_message(embed=user_card.embed_frame,
                                                                        view=user_card.view_frame)
        elif user_card.image != None:
            user_card.embed_frame = Embed(title="")
            user_card.embed_frame.set_image(url="attachment://GameImage.png")
            await user_card._buffer_image()
            user_card.message = await interaction.response.edit_message(embed=user_card.embed_frame,
                                                                        view=user_card.view_frame,
                                                                        attachments=[user_card.image_file])
        else:
            _.logger.info("Dashboard has nothing on it.")


    async def home(_, user_card:Card, interaction:Interaction) -> None:
        '''
        Brings the user back to the entry() function card
        '''

        try:
            await _.entry(user_card)
        except Exception as e:
            _.logger.info(f"Exception: {e}")

        await user_card._construct()
        if user_card.view_frame.total_children_count > 0 and user_card.image == None:
            user_card.message = await interaction.response.edit_message(embed=user_card.embed_frame,
                                                                        view=user_card.view_frame)
        elif user_card.image != None:
            user_card.embed_frame = Embed(title="")
            user_card.embed_frame.set_image(url="attachment://GameImage.png")
            await user_card._buffer_image()
            user_card.message = await interaction.response.edit_message(embed=user_card.embed_frame,
                                                                        view=user_card.view_frame,
                                                                        attachments=[user_card.image_file])
        else:
            _.logger.info("Dashboard has nothing on it.")


class Color(tuple):
    __slots__ = []
    def __new__(Class,r,g,b,a=255):
        return super().__new__(Class,[r,g,b,a])

    @property
    def r(_): return _[0]
    @property
    def g(_): return _[1]
    @property
    def b(_): return _[2]
    @property
    def a(_):return _[3]

    def __repr__(_):
        return f"Color({_.r},{_.g},{_.b},{_.a})"


WHITE = Color(255,255,255,255)
BLACK = Color(0,0,0,255)
GRAY = Color(30,30,30,255)
DEBUG_COLOR = Color(255,0,255,255)
TRANSPRENCY = Color(0,0,0,0)


from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING: from cord import Cord

from decimal import Decimal, InvalidOperation

from .component import *
from .board_item import BoardItem

from ..utilities import format_numeric


class Board(Component):
    def __init__(_, cord:Cord, x:int, y:int, parent:Component,
                 width:int|None, height:int|None,
                 items:list[str], font:Font, separation:int,
                 horizontal:bool, vertical_center:bool, horizontal_center:bool) -> None:
        super().__init__(cord=cord, x=x, y=y, parent=parent, width=width, height=height)
        _.font = font if font is not None else _.font
        _.height = _.cord.height
        _.items = items
        _.separation = separation
        _.horizontal = horizontal
        _.vertical_center = vertical_center
        _.horizontal_center = horizontal_center


    async def draw(_) -> Image:
        await super().draw()
        if _.border:
            _.drawing.rectangle([0, 0, _.width-1, _.height-1], outline=_.border_color, width=_.border_width)
            
        total_height = sum(max((item.font.height if item.font else _.font.height),(item.image.height if item.image else 0)) + _.separation for item in _.items)
        if _.vertical_center:
            y = _.x_center - (total_height // 2) if _.vertical_center else _.y
        else:
            y = (total_height // 2) if _.vertical_center else _.y
        ruler = 0
        item:BoardItem
        for item in _.items:
            font = item.font if item.font is not None else _.font
            numeric = None
            try:numeric = await format_numeric(float(Decimal(item.text.replace(",",""))))
            except InvalidOperation: pass
            font_width = await _.get_text_width(numeric, font=font) if numeric else await _.get_text_width(item.text, font=font)
            if item.image:
                if _.horizontal_center:
                    image_x = _.x_center - font_width//2 - item.image.width + item.separation
                else:
                    image_x = font_width//2 - item.image.width + item.separation
                image_y = y + ruler
                _.image.paste(im=item.image, box=(image_x, image_y), mask=item.image)
            if _.horizontal_center:
                text_x = _.x_center - font_width//2 + ((item.image.width + item.separation)//2 if item.image else 0)
            else:
                text_x = font_width//2 + ((item.image.width + item.separation)//2 if item.image else 0)
            _.drawing.text((text_x, y + ruler),
                          numeric if numeric else item.text,
                          font=font.font,
                          fill=WHITE)
            ruler += font.height + _.separation
        return _.image


from subprocess import Popen
from CordForge.launcher import Launcher

def run():
    print("Launcher running...")
    try:
        Launcher()
    except Exception as e:
        print(e)
    print("Launcher closing...")

if __name__ == "__main__":
    run()


from discord import File as DiscordFile
from discord import ButtonStyle, Embed, Intents, Member, Interaction, Message
from discord.ui import Button, View
from discord.ext.commands import Context
from PIL import Image
from io import BytesIO
from typing import Callable, Any

from .components import *
from .colors import *
from .vector2 import Vector2
from .font import Font
from .user import User

class Card:
    def __init__(_, user:User=None) -> None:
        _.user = user
        _.view_frame:View = None
        _.embed_frame:Embed = None
        _.image:Image = None
        _.image_components:list[Component] = []
        _.image_file:DiscordFile = None
        _.view_content = []
        _.embed_content = []
        _.dashboard_background = GRAY
        _.height = 640
        _.width = 640
        _.font = Font(24)
        _.message:Message = None


    async def _construct(_) -> "Card":
        await _._construct_view()
        await _._construct_components()
        return _


    async def _construct_components(_):
        image_component:Component
        for image_component in _.image_components:
            component_image:Image = await image_component.draw()
            _.image.paste(im=component_image, box=(image_component.x, image_component.y), mask=component_image.split()[3])
        _.image_components = []


    async def _construct_view(_) -> None:
        _.view_frame = View(timeout=144000)
        if len(_.view_content) > 0:
            for content in _.view_content:
                _.view_frame.add_item(content)
        _.view_content = []
    
    
    async def _buffer_image(_) -> DiscordFile:
        buffer = BytesIO()
        _.image.save(buffer, format="PNG")
        buffer.seek(0)
        _.image_file = DiscordFile(buffer, filename="GameImage.png")
        buffer.close()
        return _.image_file


    async def new_image(_) -> Image:
        'Create new Image.'
        _.image = Image.new("RGBA",
                            (_.height, _.width),
                            color=_.dashboard_background)
        return _.image
    

    async def save_image(_, path:str="CordImage") -> None:
        'Save an image to file.'
        if not hasattr(_, "image") or _.image is None:
            raise ValueError("No image found. Did you run Create_Image first?")
        _.image.save(path + ".PNG", format="PNG")


    async def add_button(_, label:str, callback:Callable, arguments:list) -> None:
        'Adds a Discord Button to the Card\'s View.'
        new_button = Button(label=label, style=ButtonStyle.grey)
        new_button.callback = lambda interaction: callback(_, interaction, *arguments)
        _.view_content.append(new_button)
    

    async def panel(_, x:int=0, y:int=0, parent:Component|None=None,
                    width:int|None=None, height:int|None=None, 
                    background:Color=GRAY, border:bool=False) -> Component:
        '''
        Create a Container Component\n
        A container's height and width is by default the parent container if given one, elsewise it's the Cord object that is it is created with.
        '''
        new_container = Panel(cord=_, x=x, y=y, parent=parent,
                                width=width, height=height,
                                background=background, border=border)
        if parent == None:
            _.image_components.append(new_container)
        else:
            parent.children.append(new_container)
        return new_container


    async def line(_, x:int=0, y:int=0, parent:Component|None=None,
                   start:Vector2=Vector2(0,0), end:Vector2=Vector2(0,0),
                   color:Color=WHITE, fill_width:int=1,
                   curve:bool=False) -> None:
        '''
        Draw a line onto the Card's Image
        '''
        new_line = Line(cord=_, x=x, y=y, parent=parent,
                       start=start, end=end,
                       fill_width=fill_width, color=color, curve=curve)
        if parent == None:
            _.image_components.append(new_line)
        else:
            parent.children.append(new_line)
        return new_line


    async def board(_, x:int=0, y:int=0, parent:Component|None=None,
                   width:int|None=None, height:int|None=None,
                   items:list[str:board_item] = [], font=None,
                   separation:int=4, horizontal:bool=False,
                   vertical_center:bool=False, horizontal_center:bool=False) -> None:
        '''
        Draw a Board onto the Card's Image
        '''
        new_display = Board(cord=_, x=x, y=y, parent=parent,
                       width=width, height=height,
                       items=items, font=font,
                       separation=separation,
                       horizontal=horizontal, vertical_center=vertical_center,
                       horizontal_center=horizontal_center)
        if parent == None:
            _.image_components.append(new_display)
        else:
            parent.children.append(new_display)
        return new_display
    

    async def text(_, content, position:list[int,int]|Vector2|None=None, parent:Component|None=None,
                   color:Color=WHITE, background:Color=None, font:Font=None,
                   center:bool=False) -> Component:
        '''
        Draw text onto the Card's Image
        '''
        new_text = Text(cord=_, content=content, position=position, parent=parent, color=color, background=background, font=font, center=center)
        if parent == None:
            _.image_components.append(new_text)
        else:
            parent.children.append(new_text)
        return new_text
    

    async def sprite(_, x:int=0, y:int=0, parent:Component|None=None,
                    sprite_image:Image=None, path:str=None) -> None:
        '''
        Copy another image ontothe Card's Image
        '''
        new_sprite = Sprite(cord=_, x=x, y=y, parent=parent, sprite_image=sprite_image, path=path)
        if parent == None:
            _.image_components.append(new_sprite)
        else:
            parent.children.append(new_sprite)
        return new_sprite


    async def debug(_, vertical_center:bool=False, horizontal_center:bool=False) -> None:
        '''
        Draw debugging drawings like vertical, and horizontal, lines.
        '''
        if vertical_center:
            await _.line(start=Vector2(_.x_center, 0), end=Vector2(_.x_center, _.height), fill_width=3, color=DEBUG_COLOR)
        if horizontal_center:
            await _.line(start=Vector2(0, _.y_center), end=Vector2(_.width, _.y_center), fill_width=3, color=DEBUG_COLOR)


from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING: from .cord import Cord

from asyncio import sleep
from os.path import exists, join
from os import mkdir, listdir, remove

from discord import Member
from .user import User
from .object import Object

users_directory = join("data", "users")

class Data:
    cord:Cord
    autosave_interval:int
    def __init__(_, cord:Cord):
        object.__setattr__(_, "cord", cord)
        _.autosave_interval = 15
        if not exists("data"):
            mkdir("data")
        if not exists(users_directory):
            mkdir(users_directory)


    def __setattr__(_, name, value):
        if name == "cord":
            raise AttributeError(f"Cannot modify Data.cord.")
        
        if isinstance(value, dict) or name in ["autosave_interval"]:
            super().__setattr__(name, value)
        else:
            raise AttributeError(f"Data attributes can only be dictonaries")


    async def autosave(_) -> None:
        print("Starting autosave loop...")
        while True:
            await sleep(_.autosave_interval)
            print("Autosaving...")
            user:User
            for user in _.cord.user_profiles.values():
                with open(join(users_directory, f"{user.id}.cf"), "w") as file:
                    data_string = ""
                    for name, value in user.data.items():
                        data_string += f"{name}={value}\n"
                    data_string = data_string[:-1]
                    file.write(data_string)

            name:str
            object:Object
            with open(join("data", f"objects.cf"), "w") as file:
                for object_name, object in _.cord.objects.items():
                    data_string = ""
                    data_string += f"{object_name}:"
                    for name, value in object.data.items():
                        data_string += f"{name}={value}~"
                    data_string = data_string[:-1] + "\n"
                    file.write(data_string)

            name:str
            data_dict:dict
            for name, data_dict in _.__dict__.items():
                if name not in ["cord", "autosave_interval"]:
                    with open(join("Data", f"{name}.cf")) as file:
                        data_string = ""
                        for name, value in data_dict.items():
                            data_string += f"{name}={value}\n"
                        data_string = data_string[:-1]
                        file.write(data_string)
            print("Finished autosaving")


    async def load_data(_) -> None:
        print("Loading data...")
        for file in listdir(users_directory):
            print(file)
            id = int(file[:-3])
            with open(join(users_directory, file), 'r') as file:
                contents = [line.strip() for line in file.readlines() if line != ""]
                for guild in _.cord.guilds:
                    member = guild.get_member(id)

                if member:
                    user = User(member)
                    _.cord.user_profiles.update({id:user})
                    for line in contents:
                        name, value = line.split("=")
                        if value.replace(".", "").isdecimal():
                            value = float(value)
                        elif value.isdecimal():
                            value = int(value)
                        user.__setattr__(name, value)
                    print(f"Loaded {member.name}'s Data")
        print("All data loaded")


    async def reset_user(_, user:User) -> None:
        user_file_path = join(users_directory, f"{user.id}.cf")
        if exists(user_file_path):
            remove(user_file_path)
            print("Reset User")
        else:
            print("Tried to reset a user's file that did not exist.")


class Vector2:
    def __init__(_, x: int = 0, y: int = 0):
        if not isinstance(x, int) or not isinstance(y, int):
            raise TypeError("Vector2 only supports integers.")
        _.x = x
        _.y = y

    def __add__(_, other: "Vector2") -> "Vector2":
        return Vector2(_.x + other.x, _.y + other.y)

    def __sub__(_, other: "Vector2") -> "Vector2":
        return Vector2(_.x - other.x, _.y - other.y)

    def __mul__(_, value: int) -> "Vector2":
        if not isinstance(value, int):
            raise TypeError("Multiplication only supports integers.")
        return Vector2(_.x * value, _.y * value)

    def __floordiv__(_, value: int) -> "Vector2":
        if value == 0:
            raise ZeroDivisionError("Cannot divide by zero.")
        if not isinstance(value, int):
            raise TypeError("Division only supports integers.")
        return Vector2(_.x // value, _.y // value)

    def __eq__(_, other: object) -> bool:
        if not isinstance(other, Vector2):
            return False
        return _.x == other.x and _.y == other.y

    def __iter__(_) -> iter:
        return iter((_.x, _.y))

    def __repr__(_) -> str:
        return f"Vector2({_.x}, {_.y})"

    def copy(_) -> "Vector2":
        return Vector2(_.x, _.y)



from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING: from cord import Cord
from .component import *


class Sprite(Component):
    def __init__(_, cord:Cord, x:int, y:int, parent:Component,
                 sprite_image:Image, path:str) -> None:
        super().__init__(cord=cord, x=x, y=y, parent=parent,
                         width=None, height=None)
        _.sprite_image = sprite_image
        _.path = path
        if path and sprite_image is None:
            _.sprite_image = Image.open(path)


    async def draw(_) -> Image:
        _.image = Image.new("RGBA", (_.width, _.height), color=_.background)
        _.image.paste(im=_.sprite_image, box=(_.x, _.y), mask=_.sprite_image.split()[3])
        return _.image


from PIL import ImageFont

class Font:
    def __init__(_, font_path:str|None=None, size:int=24) -> None:
        _.size = size
        if isinstance(font_path, str):
            _.font = ImageFont.truetype(font_path, size)
        else:
            _.font = ImageFont.load_default(_.size)
        _.ascent, _.descent = _.font.getmetrics()
        _.height = _.ascent + _.descent


from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING: from cord import Cord

from .component import *


class Line(Component):
    def __init__(_, cord:Cord, x, y, start:Vector2, end:Vector2, parent:Component, fill_width:int, color:Color, curve:bool):
        super().__init__(cord=cord, x=x, y=y, parent=parent)
        _.start = start
        _.end = end
        _.fill_width = fill_width
        _.color = color
        _.curve = curve
    
    
    async def draw(_) -> Image:
        await super().draw()
        _.drawing.line(xy=((_.start.x, _.start.y), (_.end.x, _.end.y)),
                     fill=_.color,
                     width=_.fill_width,
                     joint="curve" if _.curve else None)
        return _.image


from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING: from cord import Cord

from .component import *


class Panel(Component):
    def __init__(_, cord:Cord, x:int, y:int, parent:Component,
                 width:int, height:int, background:Color,
                 border:bool):
        super().__init__(cord=cord, x=x, y=y, width=width, height=height, parent=parent, background=background, border=border)


    async def draw(_) -> Image:
        await super().draw()
        if _.border:
            _.drawing.rectangle([0, 0, _.width-1, _.height-1], outline=_.border_color, width=_.border_width)
        return _.image


from discord import Member
from .object import Object
from typing import Any


class User(Object):
    account:Member
    id:int
    name:str
    nickname:str

    def __init__(_, account:Member) -> None:
        super().__init__()
        _.account = account
        _.id = account.id
        _.name = account.name
        _.nickname = account.nick
        _._immutables.append["account", "id", "name"]
        _._builtins = ["nickname", "add_trait"].append(_._immutables)


    def __setattr__(_, name, value):
        if name in _._immutables:
            raise AttributeError(f"Cannot modify Player.{name}. These are determined by the user's Discord profile,\
                                 and are used by CordForge for various validations, and utilities.")
        super().__setattr__(name, value)
        _.data.update({name:value})


    @staticmethod
    def add_trait(trait_name:str, value:Any) -> None:    
        if not hasattr(User, trait_name):
            print(f"Adding trait {trait_name}")
            setattr(User, trait_name, value)


from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING: from cord import Cord
from .component import *


class Text(Component):
    def __init__(_, cord:Cord, position:list|Vector2|None, parent:Component,
                 content:str, color:Color, background:Color,
                 font:Font, center:bool):
        super().__init__(cord=cord, parent=parent, width=None, height=None, color=color, font=font, background=background)
        _.font = font if font is not None else _.font
        _.content = content
        _.center = center
        if type(position) is list:
            _.position = Vector2(position[0], position[1])
        else:
            _.position = position
        _.color = color


    async def draw(_) -> Image:
        await super().draw()
        if _.center:
            _.content_width = await _.get_text_width(_.content)
            if _.position != None:
                raise("Text component cannot be given a position, and be centered.")
            _.position = Vector2()
            if _.parent:
                _.position.x = _.parent.width//2 - _.content_width//2
                _.position.y = _.parent.height//2 - _.font.height//2
            else:
                _.position.x = _.cord.width//2 - _.content_width//2
                _.position.y = _.cord.height//2 - _.font.height//2
        _.drawing.text(text=_.content,
                     xy=(_.position.x, _.position.y),
                     fill=_.color,
                     font=_.font.font)
        return _.image


class Object:
    data:dict
    _immutables:list[str]
    _builtsin:list[str]
    def __init__(_) -> None:
        object.__setattr__(_, "_immutables", [])
        object.__setattr__(_, "_builtins", [])
        object.__setattr__(_, "data", {})

        for key, value in Object.__dict__.items():
            if not key.startswith("__") and key not in _._builtins:
                _.data[key] = value


    def __setattr__(_, name, value):
        if name in _._immutables:
            raise AttributeError(f"Cannot modify {name}.")
        super().__setattr__(name, value)
        _.data.update({name:value})